/* AutoNOSim */

#ifndef __AUTONOSIM__
#define __AUTONOSIM__

#ifdef __cplusplus
extern "C" {
#endif



/* Initialization and Termination */

void AutoNOSim_Initialize (int *argc, char ***argv);

void AutoNOSim_Terminate (void);

int AutoNOSim_NTasks (void);

int AutoNOSim_MyTaskId (void);

#define AUTONOSIM_MASTER_TASK_ID 0



/* General Communication */

#define AUTONOSIM_PACKING 1
#define AUTONOSIM_UNPACKING 0

void AutoNOSim_Clear (int isPacking);

int AutoNOSim_IsPacking (void);

void AutoNOSim_Integer (int *value);

void AutoNOSim_Integers (int values[], int size);

void AutoNOSim_Real (double *value);

void AutoNOSim_Reals (double values[], int size);

void AutoNOSim_String (char values[], int size);

void AutoNOSim_SendTo (int taskId);

void AutoNOSim_ReceiveFrom (int taskId);

int AutoNOSim_ReceiveFromAny (void);

void AutoNOSim_SendReceive (int taskId);

void AutoNOSim_Barrier (void);

void AutoNOSim_GetSum 
(double *sum_IO);

void AutoNOSim_GetSum2
(double *sum0_IO, double *sum1_IO);

void AutoNOSim_GetMin 
(double *min_IO);

void AutoNOSim_GetMax 
(double *max_IO);



/* Data Exchange Across Domain Boundary */

void AutoNOSim_ReadExchangeFile (char *baseFileName);

int AutoNOSim_NEntitys ();

int AutoNOSim_NGlobalEntitys ();

int AutoNOSim_EntityIsExcluded (int iEntity);

int AutoNOSim_EntityIsOnBoundary (int iEntity);

#define AUTONOSIM_OP_SUM 0
#define AUTONOSIM_OP_PRODUCT 1
#define AUTONOSIM_OP_MIN 2
#define AUTONOSIM_OP_MAX 3

void AutoNOSim_Exchange 
(int operation,      /* AOTONOSIM_OP_* */
 int nComponents, 
 double values[]);   /* values[nEntitys][nComponents] */

void AutoNOSim_PreExchange 
(int nComponents, 
 double values[]);   /* values[nEntitys][nComponents] */

void AutoNOSim_CheckExchange (void);

void AutoNOSim_PostExchange 
(int operation,      /* AOTONOSIM_OP_* */
 int nComponents, 
 double values[]);   /* values[nEntitys][nComponents] */



/* Reduction Over The Whole Domain */

double AutoNOSim_ScalarProduct 
(int nComponents, double values0[], double values1[]);

void AutoNOSim_GetTwoScalarProducts 
(int nComponents, 
 double values0[], double values1[], double *scalarProductOfValues01, 
 double values2[], double values3[], double *scalarProductOfValues23);



#ifdef __cplusplus
}
#endif

#endif  /* __AUTONOSIM__ */

