/* autonosim_ssh.c : implementation using SSH */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <unistd.h>

#include "autonosim.h"



#if 1
#define DEBUG_MSG
#endif



/* Initialization and Termination */

static int NTasks;
static int MyTaskId;

void AutoNOSim_Initialize (int *argc, char ***argv)
{
  assert(3 <= *argc);

  NTasks = atoi ((*argv)[1]);
  MyTaskId = atoi ((*argv)[2]);

  /* NEED TO REDEFINE */
  /* modify argc and argv ??? */

#ifdef DEBUG_MSG
  fprintf (stderr,
	   " [TaskId %d] Initialize OK NTasks %d \n", 
	   MyTaskId, NTasks);
#endif
}

void AutoNOSim_Terminate (void)
{

#ifdef DEBUG_MSG
  fprintf (stderr,
	   " [TaskId %d] Terminate OK \n", 
	   MyTaskId);
#endif
}

int AutoNOSim_NTasks (void)
{
  return NTasks;
}

int AutoNOSim_MyTaskId (void)
{
  return MyTaskId;
}



/* General Communication */

#define BUFFER_SIZE 1000000

static int IsPacking;
static int PackingPosition;
static int UnpackingPosition;
static char Buffer[BUFFER_SIZE];

static void Pack (void* data, int size)
{
  assert(PackingPosition + size < BUFFER_SIZE);

  memcpy (&Buffer[PackingPosition], data, size);
  PackingPosition += size;
}

static void Unpack (void* data, int size)
{
  assert(UnpackingPosition + size < BUFFER_SIZE);

  memcpy (data, &Buffer[UnpackingPosition], size);
  UnpackingPosition += size;
}

void AutoNOSim_Clear (int isPacking)
{
  IsPacking = isPacking;
  PackingPosition = 0;
  UnpackingPosition = 0;
}

int AutoNOSim_IsPacking (void)
{
  return IsPacking;
}

void AutoNOSim_Integer (int *value)
{
  if (IsPacking) {
    Pack (value, sizeof (int));
  } else {
    Unpack (value, sizeof (int));
  }
} 

void AutoNOSim_Integers (int values[], int size)
{
  if (IsPacking) {
    Pack (values, sizeof (int) * size);
  } else {
    Unpack (values, sizeof (int) * size);
  }
} 

void AutoNOSim_Real (double *value)
{
  if (IsPacking) {
    Pack (value, sizeof (double));
  } else {
    Unpack (value, sizeof (double));
  }
}

void AutoNOSim_Reals (double values[], int size)
{
  if (IsPacking) {
    Pack (values, sizeof (double) * size);
  } else {
    Unpack (values, sizeof (double) * size);
  }
}

void AutoNOSim_String (char values[], int size)
{
  if (IsPacking) {
    Pack (values, sizeof (char) * size);
  } else {
    Unpack (values, sizeof (char) * size);
  }
}

void AutoNOSim_SendTo (int taskId)
{
  if (taskId != MyTaskId) {
    char fileName[256];
    FILE *fp;

    sprintf (fileName, "from%dto%d", MyTaskId, taskId);
    fp = fopen (fileName, "wb");
    assert(fp);

    fwrite (&PackingPosition, sizeof (int), 1, fp);
    fwrite (Buffer, PackingPosition, 1, fp);

    fclose (fp);
  }
}

void AutoNOSim_ReceiveFrom (int taskId)
{
  if (taskId != MyTaskId) {
    char fileName[256];
    FILE *fp;
    int size;
    
    sprintf (fileName, "from%dto%d", taskId, MyTaskId);
    for (;;) {
      fp = fopen (fileName, "rb");
      if (fp != NULL) {
	break;
      }
      
      sleep (1);
    }
    
    fread (&size, sizeof (int), 1, fp);
    assert(size <= BUFFER_SIZE);
    fread (Buffer, size, 1, fp);

    fclose (fp);

    remove (fileName);
  }
}  

int AutoNOSim_ReceiveFromAny (void)
{
  int taskIdFrom;

  for (;;) {
    int iTask;

    taskIdFrom = -1;
    for (iTask = 0; iTask < NTasks; iTask++) {
      char fileName[256];
      FILE *fp;
      
      if (iTask == MyTaskId) continue;
      
      sprintf (fileName, "from%dto%d", iTask, MyTaskId);
      fp = fopen (fileName, "rb");
      if (fp != NULL) {
	int size;
	
	fread (&size, sizeof (int), 1, fp);
	assert(size <= BUFFER_SIZE);
	fread (Buffer, size, 1, fp);
	
	fclose (fp);
	
	remove (fileName);
	
	taskIdFrom = iTask;
	break;
      }
    }
    if (0 <= taskIdFrom) {
      break;
    }
    
    sleep (1);
  }

  return taskIdFrom;
}  

void AutoNOSim_Barrier (void)
{
  if (MyTaskId == AUTONOSIM_MASTER_TASK_ID) {

    for (;;) {
      int count;
      int iTask;

      count = 0;
      for (iTask = 1; iTask < NTasks; iTask++) {
	char fileName[256];
	FILE *fp;
	
	sprintf (fileName, "barrier%d", iTask);
	fp = fopen (fileName, "r");
	if (fp != NULL) {
	  fclose (fp);
	  count++;
	}
      }
      if (count == NTasks - 1) {
	for (iTask = 1; iTask < NTasks; iTask++) {
	  char fileName[256];
	  
	  sprintf (fileName, "barrier%d", iTask);
	  remove (fileName);
	}
	break;
      }

      sleep (1);
    }

  } else {
    char fileName[256];
    FILE *fp;

    sprintf (fileName, "barrier%d", MyTaskId);
    fp = fopen (fileName, "w");
    assert(fp);
    fclose (fp);
  }
}  
void AutoNOSim_GetSum 
(double *sum_IO)
{
  double result;
  double domainSum;

  domainSum = *sum_IO;

  /* 
  MPI_Allreduce (&domainSum, &result, 1, 
		 MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  */

  *sum_IO = result;
}

void AutoNOSim_GetSum2
(double *sum0_IO, double *sum1_IO)
{
  double results[2];
  double domainSums[2];

  domainSums[0] = *sum0_IO;
  domainSums[1] = *sum1_IO;

  /*
  MPI_Allreduce (domainSums, results, 2, 
		 MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  */

  *sum0_IO = results[0];
  *sum1_IO = results[1];
}



/* Data Exchange Across Domain Boundary */

#define MAX_COMPONENTS 6
#define MAX_ENTITYS 1000000
#define MAX_BOUNDARYS 50
#define MAX_BOUNDARY_ENTITYS 10000

static int NEntitys;
static int NGlobalEntitys;

static int Entity_IsExcluded[MAX_ENTITYS];

static int Entity_IsOnBoundary[MAX_ENTITYS];

static int NBoundarys;

static int Boundary_NeighborDomainIds[MAX_BOUNDARYS];

static int Boundary_NEntitys[MAX_BOUNDARYS];

static int Boundary_EntityIds[MAX_BOUNDARYS][MAX_BOUNDARY_ENTITYS];
/* Domain-local entity ID */
/* They are sorted by their corresponding global entity ID. */

static double Boundary_ValuesToBeSent
[MAX_BOUNDARYS][MAX_BOUNDARY_ENTITYS * MAX_COMPONENTS];
static double Boundary_ValuesReceived
[MAX_BOUNDARYS][MAX_BOUNDARY_ENTITYS * MAX_COMPONENTS];
/* Send-receive buffer of exchanged entity values */

/*
static MPI_Request Requests[MAX_BOUNDARYS * 2];
*/

static int RequestIsDone[MAX_BOUNDARYS * 2];

void AutoNOSim_ReadExchangeFile (char *baseFileName)
{
  int domainId;
  char fileName[256];
  FILE* fp;
  int nExcludedEntitys;
  int i;
  int iBoundary;
  int iEntity;
  
  domainId = MyTaskId;
  
  sprintf (fileName, "%s.%d", baseFileName, domainId);
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", &NEntitys);
  fscanf (fp, "%d", &NGlobalEntitys);

  for (iEntity = 0; iEntity < NEntitys; iEntity++) {
    Entity_IsExcluded[iEntity] = 0;
  }

  fscanf (fp, "%d", &nExcludedEntitys);
  
  for (i = 0; i < nExcludedEntitys; i++) {
    int entityId;

    fscanf (fp, "%d", &entityId);
    Entity_IsExcluded[entityId] = 1;
  }

  fscanf (fp, "%d", &NBoundarys);
  
  for (iBoundary = 0; iBoundary < NBoundarys; iBoundary++) {
    fscanf (fp, "%d", &Boundary_NeighborDomainIds[iBoundary]);
    fscanf (fp, "%d", &Boundary_NEntitys[iBoundary]);
    
    for (iEntity = 0; 
	 iEntity < Boundary_NEntitys[iBoundary]; 
	 iEntity++) {
      fscanf (fp, "%d", &Boundary_EntityIds[iBoundary][iEntity]);
    }
  }
  
  fclose (fp);

  for (iEntity = 0; iEntity < NEntitys; iEntity++) {
    Entity_IsOnBoundary[iEntity] = 0;
  }

  for (iBoundary = 0; iBoundary < NBoundarys; iBoundary++) {
    
    for (iEntity = 0; 
	 iEntity < Boundary_NEntitys[iBoundary]; 
	 iEntity++) {
      int entityId = Boundary_EntityIds[iBoundary][iEntity];

      Entity_IsOnBoundary[entityId] = 1;
    }
  }
}

int AutoNOSim_EntityIsExcluded (int iEntity)
{
  assert(0 <= iEntity);
  assert(iEntity < NEntitys);
  
  return Entity_IsExcluded[iEntity];
}

int AutoNOSim_EntityIsOnBoundary (int iEntity)
{
  assert(0 <= iEntity);
  assert(iEntity < NEntitys);
  
  return Entity_IsOnBoundary[iEntity];
}

static void PrepareSendingValues (int nComponents, double values[])
{
  int iBoundary;
  int iEntity;
  int iComponent;

  for (iBoundary = 0; iBoundary < NBoundarys; iBoundary++) {

    for (iEntity = 0; 
	 iEntity < Boundary_NEntitys[iBoundary]; 
	 iEntity++) {
      int entityId = Boundary_EntityIds[iBoundary][iEntity];
      
      for (iComponent = 0; iComponent < nComponents; iComponent++) {
	Boundary_ValuesToBeSent
	  [iBoundary][iEntity * nComponents + iComponent] 
	  = values[entityId * nComponents + iComponent];
      }
    }
  }
}

static void SendReceiveBoundaryValues (int nComponents)
{
  int iBoundary;
  int iRequest;
  int count;
  
  count = 0;
  for (iBoundary = 0; iBoundary < NBoundarys; iBoundary++) {
    int destTaskId = Boundary_NeighborDomainIds[iBoundary];
    int nEntitys = Boundary_NEntitys[iBoundary];
    
    /* 
    MPI_Isend (Boundary_ValuesToBeSent[iBoundary], 
	       nEntitys * nComponents, MPI_DOUBLE, 
	       destTaskId, 0, MPI_COMM_WORLD, 
	       &Requests[count]);
    */
    count++;
    /* 
    MPI_Irecv (Boundary_ValuesReceived[iBoundary], 
	       nEntitys * nComponents, MPI_DOUBLE, 
	       destTaskId, 0, MPI_COMM_WORLD, 
	       &Requests[count]);
    */
    count++;
  }

  for (iRequest = 0; iRequest < NBoundarys * 2; iRequest++) {
    RequestIsDone[iRequest] = 0;
  }
}

static void CheckRequests ()
{
  int iRequest;
  int flag;
  /*
  MPI_Status status;
  */

  for (iRequest = 0; iRequest < NBoundarys * 2; iRequest++) {
    if (RequestIsDone[iRequest]) continue;

    flag = 0;
    /* 
    MPI_Test (&Requests[iRequest], &flag, &status);
    */
    if (flag) {
      RequestIsDone[iRequest] = 1;
    }
  }
}

static void WaitRequests ()
{
  /*
  static MPI_Status statuss[MAX_BOUNDARYS * 2];

  MPI_Waitall (NBoundarys * 2, Requests, statuss);
  */
}

static void PerformOperationOnValues 
(int operation, int nComponents, double values[])
{
  int iBoundary;
  int iEntity;
  int iComponent;

  for (iBoundary = 0; iBoundary < NBoundarys; iBoundary++) {
    
    for (iEntity = 0; 
	 iEntity < Boundary_NEntitys[iBoundary]; 
	 iEntity++) {
      int entityId = Boundary_EntityIds[iBoundary][iEntity];
      
      for (iComponent = 0; iComponent < nComponents; iComponent++) {
	double valueReceived = Boundary_ValuesReceived
	  [iBoundary][iEntity * nComponents + iComponent];
	
	switch (operation) {
	case AUTONOSIM_OP_SUM:
	  values[entityId * nComponents + iComponent] += valueReceived;
	  break;
	case AUTONOSIM_OP_PRODUCT:
	  values[entityId * nComponents + iComponent] *= valueReceived;
	  break;
	case AUTONOSIM_OP_MIN:
	  {
	    double value = values[entityId * nComponents + iComponent];
	    
	    values[entityId * nComponents + iComponent] 
	      = (value < valueReceived) ? value : valueReceived;
	  }
	  break;
	case AUTONOSIM_OP_MAX:
	  {
	    double value = values[entityId * nComponents + iComponent];
	    
	    values[entityId * nComponents + iComponent] 
	      = (value < valueReceived) ? valueReceived : value;
	  }
	  break;
	default:
	  assert(0);
	  break;
	}
      }
    }
  }
}  

void AutoNOSim_Exchange 
(int operation, int nComponents, double values[])
{
  assert(nComponents <= MAX_COMPONENTS);

  PrepareSendingValues (nComponents, values);
  SendReceiveBoundaryValues (nComponents);
  WaitRequests ();
  PerformOperationOnValues (operation, nComponents, values);

#ifdef DEBUG_MSG
  fprintf (stderr,
	   " [TaskId %d] exchange ( operationID %d ) \n", 
	   MyTaskId, operation);
#endif
}

void AutoNOSim_PreExchange 
(int nComponents, double values[])
{
  assert(nComponents <= MAX_COMPONENTS);

  PrepareSendingValues (nComponents, values);
  SendReceiveBoundaryValues (nComponents);

#ifdef DEBUG_MSG
  fprintf (stderr,
	   " [TaskId %d] pre-exchange \n", 
	   MyTaskId);
#endif
}

void AutoNOSim_CheckExchange (void)
{
  CheckRequests ();
}

void AutoNOSim_PostExchange 
(int operation, int nComponents, double values[])
{
  assert(nComponents <= MAX_COMPONENTS);

  WaitRequests ();
  PerformOperationOnValues (operation, nComponents, values);

#ifdef DEBUG_MSG
  fprintf (stderr,
	   " [TaskId %d] post-exchange ( operationID %d ) \n", 
	   MyTaskId, operation);
#endif
}

double AutoNOSim_ScalarProduct 
(int nComponents, double values0[], double values1[])
{
  double result;
  double domainSum;
  int iEntity;
  int iComponent;

  domainSum = 0.0;
  for (iEntity = 0; iEntity < NEntitys; iEntity++) {
    if (!Entity_IsExcluded[iEntity]) {
      
      for (iComponent = 0; iComponent < nComponents; iComponent++) {
	domainSum += values0[iEntity * nComponents + iComponent]
	  * values1[iEntity * nComponents + iComponent];
      }
    }
  }

  /*
  MPI_Allreduce (&domainSum, &result, 1, 
		 MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  */

#ifdef DEBUG_MSG
  fprintf (stderr,
	   " [TaskId %d] scalar product \n", 
	   MyTaskId);
#endif

  return result;
}

void AutoNOSim_GetTwoScalarProducts 
(int nComponents, 
 double values0[], double values1[], double *scalarProductOfValues01, 
 double values2[], double values3[], double *scalarProductOfValues23)
{
  *scalarProductOfValues01 
    = AutoNOSim_ScalarProduct (nComponents, values0, values1);
  *scalarProductOfValues23 
    = AutoNOSim_ScalarProduct (nComponents, values2, values3);
}
