#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <unistd.h>

#include <autonosim.h>



int NTasks;
int MyTaskId;

#define COMMAND_EXIT 0
#define COMMAND_DO_SOMETHING 1

void Master (void)
{
  int nJobs;
  int nSlaves;
  int iSlave;
  int iJob;
  int jobCount;

  nJobs = 10;
  nSlaves = NTasks - 1;
  jobCount = 0;

  for (iSlave = 0; iSlave < nSlaves; iSlave++) {
    int slaveTaskId = iSlave + 1;
    int commandId = COMMAND_DO_SOMETHING;

    AutoNOSim_Clear (AUTONOSIM_PACKING);
    AutoNOSim_Integer (&commandId);
    AutoNOSim_Integer (&jobCount);
    AutoNOSim_SendTo (slaveTaskId);

    jobCount++;
  }

  for (iJob = 0; iJob < nJobs; iJob++) {
    int commandId = COMMAND_DO_SOMETHING;
    int slaveTaskId;
    int returnValue;
    int slaveId;

    slaveTaskId = AutoNOSim_ReceiveFromAny ();
    AutoNOSim_Clear (AUTONOSIM_UNPACKING);
    AutoNOSim_Integer (&returnValue);

    slaveId = slaveTaskId - 1;
    fprintf (stderr,
	     " [Master] return %d from slaveId %d \n", 
	    returnValue, slaveId);
    fprintf (stderr,
	     " [Master] %d jobs completed \n", 
	    iJob + 1);

    if (iJob < nJobs - nSlaves) {
      AutoNOSim_Clear (AUTONOSIM_PACKING);
      AutoNOSim_Integer (&commandId);
      AutoNOSim_Integer (&jobCount);
      AutoNOSim_SendTo (slaveTaskId);

      jobCount++;
    }
  }

  for (iSlave = 0; iSlave < nSlaves; iSlave++) {
    int commandId = COMMAND_EXIT;
    int slaveTaskId = iSlave + 1;
    
    AutoNOSim_Clear (AUTONOSIM_PACKING);
    AutoNOSim_Integer (&commandId);
    AutoNOSim_SendTo (slaveTaskId);
  }
  for (iSlave = 0; iSlave < nSlaves; iSlave++) {
    int slaveTaskId = iSlave + 1;
    int returnValue;
    
    AutoNOSim_ReceiveFrom (slaveTaskId);
    AutoNOSim_Clear (AUTONOSIM_UNPACKING);
    AutoNOSim_Integer (&returnValue);
    assert(returnValue == 0);
  }

  fprintf (stderr, 
	   " [Master] EXIT \n");
}

void Slave (void)
{
  for (;;) {
    int slaveId = MyTaskId - 1;
    int commandId;

    AutoNOSim_ReceiveFrom (AUTONOSIM_MASTER_TASK_ID);
    AutoNOSim_Clear (AUTONOSIM_UNPACKING);
    AutoNOSim_Integer (&commandId);

    switch (commandId) {
    case COMMAND_EXIT:
      {
	int returnValue;

	fprintf (stderr,
		 " [SlaveId %d] Slave commandId %d EXIT \n", 
		 slaveId, commandId);

	returnValue = 0;

	AutoNOSim_Clear (AUTONOSIM_PACKING);
	AutoNOSim_Integer (&returnValue);
	AutoNOSim_SendTo (AUTONOSIM_MASTER_TASK_ID);
	
	return;
      }
      break;
    case COMMAND_DO_SOMETHING:
      {
	int jobCount;
	int returnValue;

	AutoNOSim_Integer (&jobCount);

	fprintf (stderr,
		 " [SlaveId %d] Slave commandId %d DO SOMETHING job %d \n", 
		 slaveId, commandId, jobCount);
	
	returnValue = (slaveId + 1) * 10;
	sleep (returnValue);
	
	AutoNOSim_Clear (AUTONOSIM_PACKING);
	AutoNOSim_Integer (&returnValue);
	AutoNOSim_SendTo (AUTONOSIM_MASTER_TASK_ID);
      }
      break;
    }
    
  }
}

int main (int argc, char *argv[]) 
{
  AutoNOSim_Initialize (&argc, &argv);

  NTasks = AutoNOSim_NTasks ();
  MyTaskId = AutoNOSim_MyTaskId ();
  fprintf (stderr, 
	   " <taskId %d> num. tasks %d \n", 
	   MyTaskId, NTasks);

  assert(2 <= NTasks);

  if (MyTaskId == AUTONOSIM_MASTER_TASK_ID) {
    Master ();
  } else {
    Slave ();
  }

  AutoNOSim_Terminate ();

  return 0;
}



