#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



#define MAX_GLOBAL_ENTITYS 2500000
#define MAX_DOMAINS 100
#define MAX_DOMAIN_ENTITYS 300000
#define MAX_BOUNDARYS 50
#define MAX_BOUNDARY_ENTITYS 10000



/* IN */

int NGlobalEntitys;

int NDomains;

int Domain_NEntitys[MAX_DOMAINS];

int Domain_GlobalEntityIds[MAX_DOMAINS][MAX_DOMAIN_ENTITYS];



/* OUT */

int Domain_NBoundarys[MAX_DOMAINS];

int Boundary_NeighborDomainIds[MAX_DOMAINS][MAX_BOUNDARYS];

int Boundary_NEntitys[MAX_DOMAINS][MAX_BOUNDARYS];

int Boundary_EntityIds[MAX_DOMAINS][MAX_BOUNDARYS][MAX_BOUNDARY_ENTITYS];
/* Domain-local entity ID */
/* They are sorted by their corresponding global entity ID. */

int Domain_OwnsEntity[MAX_DOMAINS][MAX_DOMAIN_ENTITYS];



/* WORK */

#define MAX_INTERDOMAIN_DOMAINS 20

int GlobalEntity_NDomains[MAX_GLOBAL_ENTITYS];

int GlobalEntity_DomainIds[MAX_GLOBAL_ENTITYS][MAX_INTERDOMAIN_DOMAINS];

int GlobalEntity_EntityIds[MAX_GLOBAL_ENTITYS][MAX_INTERDOMAIN_DOMAINS];
/* Domain-local entity ID */

int GlobalEntity_IsOwned[MAX_GLOBAL_ENTITYS];



void ReadGlobalFile (char* fileName)
{
  FILE* fp;

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", &NGlobalEntitys);
  assert(NGlobalEntitys <= MAX_GLOBAL_ENTITYS);

  fscanf (fp, "%d", &NDomains);
  assert(NDomains <= MAX_DOMAINS);

  fclose (fp);
}

void ReadDomainFile (char* baseFileName)
{
  int iDomain;

  for (iDomain = 0; 
       iDomain < NDomains; 
       iDomain++) {
    char fileName[256];
    FILE* fp;
    int iEntity;

    sprintf (fileName, "%s.%d", baseFileName, iDomain);
    fp = fopen (fileName, "r");
    assert(fp != NULL);
    
    fscanf (fp, "%d", &Domain_NEntitys[iDomain]);
    assert(Domain_NEntitys[iDomain] <= MAX_DOMAIN_ENTITYS);

    for (iEntity = 0; iEntity < Domain_NEntitys[iDomain]; iEntity++) {
      fscanf (fp, "%d", &Domain_GlobalEntityIds[iDomain][iEntity]);
      assert(0 <= Domain_GlobalEntityIds[iDomain][iEntity]);
      assert(Domain_GlobalEntityIds[iDomain][iEntity] 
	     < MAX_GLOBAL_ENTITYS);
    }

    fclose (fp);
  }
}

void InitializeInterdomainTable (void)
{
  int iGlobalEntity;

  for (iGlobalEntity = 0; 
       iGlobalEntity < NGlobalEntitys; 
       iGlobalEntity++) {
    GlobalEntity_NDomains[iGlobalEntity] = 0;
  }
}

void AssureInterdomainTable (void)
{
  int iGlobalEntity;

  for (iGlobalEntity = 0; 
       iGlobalEntity < NGlobalEntitys; 
       iGlobalEntity++) {
    /* A global entity which is not referred from any domain is found. */
    assert(1 <= GlobalEntity_NDomains[iGlobalEntity]);
  }
}

void MakeInterdomainTable (void)
{
  int iDomain;

  InitializeInterdomainTable ();

  for (iDomain = 0; 
       iDomain < NDomains; 
       iDomain++) {
    int iEntity;
    
    for (iEntity = 0; 
	 iEntity < Domain_NEntitys[iDomain]; 
	 iEntity++) {
      int globalEntityId = Domain_GlobalEntityIds[iDomain][iEntity];
      int count;
      
      assert(GlobalEntity_NDomains[globalEntityId] < MAX_INTERDOMAIN_DOMAINS);
      count = GlobalEntity_NDomains[globalEntityId];
      GlobalEntity_DomainIds[globalEntityId][count] = iDomain;
      GlobalEntity_EntityIds[globalEntityId][count] = iEntity;
      count++;
      GlobalEntity_NDomains[globalEntityId] = count;
    }    
  }

  AssureInterdomainTable ();
}

void InitializeBoundary (void)
{
  int iDomain;

  for (iDomain = 0; 
       iDomain < NDomains; 
       iDomain++) {
    Domain_NBoundarys[iDomain] = 0;
  }
}

void AddBoundaryEntity (int domainId, int entityId, int neighborDomainId)
{
  int iBoundary;
  int boundaryId;

  boundaryId = -1;
  for (iBoundary = 0; 
       iBoundary < Domain_NBoundarys[domainId]; 
       iBoundary++) {
    if (Boundary_NeighborDomainIds[domainId][iBoundary] == neighborDomainId) {
      boundaryId = iBoundary;
      break;
    }
  }

  if (boundaryId == -1) {
    int count;

    assert(Domain_NBoundarys[domainId] < MAX_BOUNDARYS);
    count = Domain_NBoundarys[domainId];
    Boundary_NeighborDomainIds[domainId][count] = neighborDomainId;
    boundaryId = count;
    count++;
    Domain_NBoundarys[domainId] = count;
  }
  assert(0 <= boundaryId);
  assert(boundaryId < Domain_NBoundarys[domainId]);
  
  {
    int count;

    assert(Boundary_NEntitys[domainId][boundaryId] < MAX_BOUNDARY_ENTITYS);
    count = Boundary_NEntitys[domainId][boundaryId];
    Boundary_EntityIds[domainId][boundaryId][count] = entityId;
    count++;
    Boundary_NEntitys[domainId][boundaryId] = count;
  }
}

void AssureBoundary (void)
{
}

void MakeBoundary (void)
{
  int iGlobalEntity;
  
  InitializeBoundary ();

  for (iGlobalEntity = 0; 
       iGlobalEntity < NGlobalEntitys; 
       iGlobalEntity++) {
    int nDomains = GlobalEntity_NDomains[iGlobalEntity];
    int iDomain;

    if (nDomains == 1) continue;
      
    for (iDomain = 0; 
	 iDomain < nDomains; 
	 iDomain++) {
      int domainId = GlobalEntity_DomainIds[iGlobalEntity][iDomain];
      int entityId = GlobalEntity_EntityIds[iGlobalEntity][iDomain];
      int jDomain;
      
      for (jDomain = 0; jDomain < nDomains; jDomain++) {
	int neighborDomainId = GlobalEntity_DomainIds[iGlobalEntity][jDomain];
	
	if (jDomain == iDomain) continue;
	
	AddBoundaryEntity (domainId, entityId, neighborDomainId);
      }
    }
  }

  AssureBoundary ();
}   

void SetEntityOwner (void)
{
  int iGlobalEntity;
  int iDomain;
  
  for (iGlobalEntity = 0; 
       iGlobalEntity < NGlobalEntitys; 
       iGlobalEntity++) {
    GlobalEntity_IsOwned[iGlobalEntity] = 0;
  }
  
  for (iDomain = 0; iDomain < NDomains; iDomain++) {
    int iEntity;
    
    for (iEntity = 0; 
	 iEntity < Domain_NEntitys[iDomain]; 
	 iEntity++) {
      int globalEntityId = Domain_GlobalEntityIds[iDomain][iEntity];
      
      if (GlobalEntity_IsOwned[globalEntityId]) {
	Domain_OwnsEntity[iDomain][iEntity] = 0;
      } else {
	Domain_OwnsEntity[iDomain][iEntity] = 1;
	GlobalEntity_IsOwned[globalEntityId] = 1;
      }
    }    
  }
}

void WriteExcludedEntity (FILE* fp, int iDomain)
{
  int iEntity;
  int count;
  
  count = 0;
  for (iEntity = 0; 
       iEntity < Domain_NEntitys[iDomain]; 
       iEntity++) {
    if (!Domain_OwnsEntity[iDomain][iEntity]) {
      count++;
    }
  }    

  fprintf (fp, "%d\n", count);
  
  fprintf (fp, "\n");
  
  for (iEntity = 0; 
       iEntity < Domain_NEntitys[iDomain]; 
       iEntity++) {
    if (!Domain_OwnsEntity[iDomain][iEntity]) {
      fprintf (fp, "%d\n", iEntity);
    }
  }    

  fprintf (fp, "\n\n");
}

void WriteBoundary (FILE* fp, int iDomain)
{
  int iBoundary;
  
  fprintf (fp, "%d\n", Domain_NBoundarys[iDomain]);
  
  fprintf (fp, "\n\n");
  
  for (iBoundary = 0; 
       iBoundary < Domain_NBoundarys[iDomain]; 
       iBoundary++) {
    int iEntity;
    
    fprintf (fp, "%d\n", Boundary_NeighborDomainIds[iDomain][iBoundary]);
    fprintf (fp, "%d\n", Boundary_NEntitys[iDomain][iBoundary]);
    
    fprintf (fp, "\n");
    
    for (iEntity = 0; 
	 iEntity < Boundary_NEntitys[iDomain][iBoundary]; 
	 iEntity++) {
      fprintf (fp, "%d\n", Boundary_EntityIds[iDomain][iBoundary][iEntity]);
    }
    
    fprintf (fp, "\n\n");
  }
}

void WriteFiles (char* baseFileName)
{
  int iDomain;
  
  for (iDomain = 0; iDomain < NDomains; iDomain++) {
    char fileName[256];
    FILE* fp;

    sprintf (fileName, "%s.%d", baseFileName, iDomain);
    fp = fopen (fileName, "w");
    assert(fp != NULL);

    fprintf (fp, "%d\n", Domain_NEntitys[iDomain]);
    fprintf (fp, "%d\n", NGlobalEntitys);
    fprintf (fp, "\n\n");

    WriteExcludedEntity (fp, iDomain);
    WriteBoundary (fp, iDomain);

    fclose (fp);
  }

}

int main (int argc, char *argv[]) 
{
  if (argc < 4) {
    fprintf 
      (stderr, 
       " Usage : make_exchange_table "
       " global_file domain_file exchange_file \n");
    exit (1);
  }

  ReadGlobalFile (argv[1]);
  ReadDomainFile (argv[2]);

  MakeInterdomainTable ();

  MakeBoundary ();

  SetEntityOwner ();

  WriteFiles (argv[3]);

  return 0;
}



